import uuid
from abc import ABC, abstractmethod
from typing import Optional

from ..types import Default, DefaultType
from ..resources import Resource, ResourceFilter


class BaseLoader(ABC):

    SLUG: str
    URI_SCHEME: str

    locations: Optional[list[str]]
    metadata: Optional[dict]
    filter: ResourceFilter

    def __init__(
            self,
            locations: list[str] | None = None,
            metadata: dict | None = None,
            exclusions: list[str] | None = None,
            filter: ResourceFilter | None = None,
        ) -> None:
        super().__init__()
        self.exclusions = exclusions or []
        self.metadata = metadata or {}
        self.filter = filter
        if locations:
            parsed_locations, parsed_exclusions = self.parse_locations(locations)
            self.locations = parsed_locations
            self.exclusions.extend(parsed_exclusions)
        else:
            self.locations = locations

    @abstractmethod
    def discover(
        self,
        locations: list[str] | DefaultType = Default,
        metadata: dict | DefaultType = Default,
        exclusions: list[str] | DefaultType = Default,
        filter: ResourceFilter | DefaultType = Default,
        *args,
        **kwargs,
    ):
        ...

    @abstractmethod
    def read(
        self,
        location: str,
        base: str = "",
    ):
        ...


    def parse_locations(
        self,
        locations: list[str],
    ) -> tuple[list[str], list[str]]:
        exclusions = []
        for location in locations[:]:
            if str(location).startswith("!"):
                locations.remove(location)
                exclusion_value = str(location).removeprefix('!')
                exclusions.append(exclusion_value)
        return locations, exclusions


    def should_exclude(
        self,
        location: str,
        exclusions: list[str] | DefaultType = Default
    ):
        if exclusions is Default:
            exclusions = self.exclusions
        for exclusion in exclusions:
            if exclusion in location:
                return True
        return False

    @classmethod
    def get_id_for_resource(cls, resource: Resource):
        # Is this check valid? Is it broken to have resource without a URI?
        # Do we need to be able to fetch it from the URI later on, even if it contains the full document content?
        if resource.uri:
            return f"{cls.SLUG}:{resource.uri}"
        else:
            return f"{cls.SLUG}:{uuid.uuid4()}"



import click
import datetime
import glob
import json
import os
import os.path
import re
import toml
from collections import defaultdict
from functools import reduce
from pathlib import Path
from typing import Any

import openai

from ..corpus.types import Default, DefaultType, URI
from ..corpus.resources import Resource
from ..corpus.loaders.schemes import unmap_scheme, read_from_uri
from ..corpus.loaders.code_library_loader import RCRANLocalCache
from ..corpus.loaders.local_file_loader import LocalFileLoader

from .helpers import find_pyproject_file, calculate_content_hash


def generate_existing_example_map(
    example_dirs: list[str],
    base_map: DefaultType | dict[str, Any] = Default,  # TODO: Fix type
):
    if base_map is Default:
        map =  {
            "examples": {},
            "sources": {},
        }
    else:
        map = base_map

    resources = LocalFileLoader(locations=example_dirs).discover()
    for resource in resources:
        source_uri = resource.metadata.get("source_uri", None)
        if not source_uri:
            # If no source_uri, it's not an example that can checked for change.
            continue
        source_hash = resource.metadata.get("source_sha256_hash", None)
        # if source_uri in map and source_hash != map[source_uri]:
        #     raise KeyError("")
        example_path = resource.uri.path
        if source_uri in map["sources"]:
            # if source_hash != map["source"][source_uri]:
            #     raise KeyError("")
            map["sources"][source_uri]["example_files"].append(example_path)
        else:
            map["sources"][source_uri] = {
                "hash": source_hash,
                "example_files": [example_path,]
            }
        map["examples"][example_path] = {
            "hash": source_hash,
            "source": source_uri,
        }

    return map


# TODO: Add optional param for setting destination
@click.command()
@click.argument("locations", type=str, required=False, nargs=-1)
def extract_examples(locations: list[str]):
    """
    Extract examples from Bunsen config or specified locations for use in RAG
    """

    dest = 'examples-unverified'
    if not os.path.exists(dest):
        os.makedirs(dest)

    sources: dict[str, list[str]] = defaultdict(lambda: [])
    if locations:
        bad_sources = []
        for location in locations:
            uri = URI(location)
            if not uri.scheme or uri.scheme == "file":
                source = Path(location)
                if not (source.is_dir() or source.is_file()):
                    bad_sources.append(source)
                else:
                    sources["file"].append(source.absolute())
        if bad_sources:
            raise click.UsageError(f"source(s) {', '.join(bad_sources)} are not valid.")
        examples_path = None
    else:
        # Get locations from config
        pyproject_file_path = find_pyproject_file()
        if pyproject_file_path:
            with pyproject_file_path.open() as pyproject_file:
                pyproject_config = toml.load(pyproject_file)
            bunsen_config = pyproject_config.get("tool", {}).get("hatch", {}).get("build", {}).get("hooks", {}).get("bunsen", {})
            sources["py-mod"].extend(
                bunsen_config.get("python_libraries", [])
            )
            sources["rcran-package"].extend(
                bunsen_config.get("r_cran_libraries", [])
            )
            documentation_path = bunsen_config.get("documentation_path", None)
            if documentation_path and Path(documentation_path).exists():
                sources["documentation"].append(
                    documentation_path
                )
            examples_path = bunsen_config.get("examples_path", None)
            # TODO: We probably shouldn't extract example from the examples, so skipping, but keeping in case we change our mind soon.
            # if examples_path and Path(examples_path).exists():
            #     sources["examples"].append(
            #         examples_path
            #         # f"examples:{examples_path}"
            #     )
        else:
            examples_path = None


        if not pyproject_file_path or not bunsen_config:
            raise click.UsageError(f"No locations provided and unable to find a bunsen configuration in the directory tree.")


    with RCRANLocalCache(locations=sources["rcran-package"]):
        resource_list: list[Resource] = []
        for scheme, scheme_locations in sources.items():
            scheme_cls = unmap_scheme(scheme)

            loader_cls = scheme_cls.default_loader()
            loader = loader_cls()
            resource_list.extend(
                loader.discover(scheme_locations)
            )

        example_locations = [dest]
        if examples_path:
            example_locations.append(examples_path)
        example_map = generate_existing_example_map(example_locations)

        existing_example_nums = [filename.split('_')[1] for filename in os.listdir(dest)]
        num = 1
        try:
            if existing_example_nums:
                num = max(map(int, existing_example_nums)) + 1
        except:
            pass
        files_to_delete = []

        for resource in resource_list:
            if resource.content:
                content = resource.content
            else:
                if resource.file_handle and not resource.file_handle.closed and resource.file_handle.readable():
                    content = resource.file_handle.read()
                else:
                    content = read_from_uri(resource.uri)

            content_hash = calculate_content_hash(content)
            if resource.uri in example_map["sources"] and example_map["sources"][resource.uri]["hash"] == content_hash:
                print(f"Hash for resource {resource.uri} has not changed for resource {resource.uri}. Still {content_hash}.\nUse --force to force extraction.")
                continue
            else:
                if resource.uri in example_map["sources"]:
                    print(f'Hash for resource {resource.uri} has changed. Was {example_map["sources"][resource.uri]["hash"]}, now {content_hash}.')
                    files_to_delete.extend(example_map["sources"][resource.uri]["example_files"])


            content_lines = content.splitlines(keepends=True)
            prompt = """
A document is provided below. It contains line numbers on the left, the content on the right, with the sides separated by ` : `.
The provided document may be a source code file, a documentation page, an executable notebook, or hand extracted examples created by hand.
It may contain one or more curated examples as for how to accomplish something using relevant libraries.
Please identify any such examples and generate a short description of the purpose of the example code, along with the start/stop line numbers (inclusive) where the example occurs.
For example, if the example is a single line on the line labeled 34, both `start_line` and `stop_line` would be equal to 34.
If the example was spread of 5 lines, starting at line 12, the values would be `start_line` = 12, `stop_line` = 16.
Also, make sure to include a prelude of code that contains any imports or definitions needed so the example is as complete as possible and has everything it needs to run as a stand-alone code block.
The prelude will later be combined with the example to help use the libraries these documents are related to.
Please ensure that all imports and definitions in the prelude are complete. Do not use statements like "put your code here..." or summarize what should be done.
Do not identify/extract regular source code from files as examples. Only extract unique tasks that demonstrate how to properly use the library.
That is, all exracted examples should show usage of functions, not just the code that defines a function. Assume the user will be able to look up argument and parameter information.
You should err on the side of fewer, more complete examples which show an entire "step" of work rather than examples of subtasks.
As an example, if you find sample code where a dataset is sorted and a comparison function is defined to help sort, only extract one example that includes all the code. Do not create a separate example of how to define a comparison function.
If you do not find any examples in the document, do not generate one yourself. Instead return an empty list.

Please be sure to format your answer in json format as response object that matches this format with one object per example object in the `examples` list:
  {
    "examples": [
      {
        "description": str,
        "prelude": str,
        "start_line": int,
        "stop_line": int,
      },
      ...
    ]
  }

The document from which to extract begins below this line and runs until the end of the input:\n"""

            full_prompt = "\n".join([
                prompt,
                "".join(f"{line_no:6} : {line}" for line_no, line in enumerate(content_lines, start=1)),
            ])
            model = "gpt-4o"

            response = openai.chat.completions.create(
                model=model,
                messages=[
                    {
                        "role": "system",
                        "content": """
    You are a helpful assistant who is excellent at understanding and describing code and who always responds using JSON.
    You are helping to identify and explain code examples to allow for better understanding and use on how to use novel and
    interesting software libaries. These examples will help users to code in and/or use the covered software.
                        """.strip(),
                    },
                    {
                        "role": "user",
                        "content": full_prompt,
                    }
                ],
                response_format={"type": "json_object"},
            )

            metadata = {
                **resource.metadata,
                "resource_id": resource.id,
                "source_uri": resource.uri,
                "generated_at": datetime.datetime.utcnow().isoformat(),
                "source_sha256_hash": content_hash,
            }

            if not response:
                continue
            response_content = response.choices[0].message.content.strip()
            if not response_content:
                continue
            try:
                example_list = json.loads(response_content)["examples"]
            except json.JSONDecodeError as err:
                click.echo(f"Encountered error {err} when processing {resource.uri}")
                failure_file_path = os.path.join(dest, f"failure_{num}.txt")
                with open(failure_file_path, "w") as failure_file:
                    failure_file.write(
                        f"{response_content}\n"
                        f"{'====' * 20}\n"
                    )
                    json.dump(metadata, failure_file, indent=2)
                    failure_file.write(
                        f"\n{'====' * 20}\n"
                        f"{err}\n{err.msg}\n"
                    )
                continue

            if not example_list:
                continue

            click.echo(f"Found {len(example_list)} examples in {resource.uri}")
            for example_num, example in enumerate(example_list, start=1):
                example_filename = os.path.join(dest, f"example_{num}_{example_num}.md")
                example_metadata_filename = f"{example_filename}.metadata"
                start = int(example["start_line"]) - 1
                stop = int(example["stop_line"])
                prelude_lines = example.get("prelude", "").splitlines(keepends=True)
                prelude = clean_codeblock(prelude_lines) if prelude_lines else ""
                example_code_lines = content_lines[start:stop]
                example_code = clean_codeblock(example_code_lines)

                example_metadata = {
                    **metadata,
                    "start_line": start,
                    "stop_line": stop,
                }

                with open(example_filename, "w") as example_file, open(example_metadata_filename, "w") as example_metadata_file:
                    example_file.write(f"""
# Description
{example["description"]}

# Code
```
{prelude}
{example_code}
```
""".lstrip())
                    json.dump(example_metadata, example_metadata_file, indent=2, default=(lambda o: str(o)))

            # Iterate source file num for next loop
            num += 1
        print("Files to delete: ", files_to_delete)


def should_ignore_line(line: str) -> bool:
    """
    Returns a boolean as to whether a particular line should be kept or ignored when importing code.
    We should ignore extraneous whitespace, code-block indicators, decorative lines, etc that do are not semantically
    meaningful.
    """
    # Todo: Other leading/trailing lines we need to clean?
    if (
        re.search(r'\S', line) == None  # Line does not contain at least one non-whitespace character
        or line.startswith('```')
    ):
        return True
    return False


def clean_codeblock(code: list[str]) -> str:
    if not code:
        return ""
    top, bottom = 0, len(code) - 1
    while should_ignore_line(code[top]) and top < len(code) - 1:
        top += 1
    while should_ignore_line(code[bottom]) and bottom > top:
        bottom -= 1
    split_lines = [re.split(r'(\s)', line) for line in code[top:bottom]]
    line_count = len(split_lines)
    vertical_parts = list(zip(*split_lines))
    for idx, section in enumerate(vertical_parts):
        correct_size = len(section) == line_count
        all_match, _ = reduce(
            lambda match_tuple, next_val: (match_tuple[1] == next_val, match_tuple[1]),
            section,
            (True, section[0])
        )
        if not (correct_size and all_match):
            break
    else:
        idx = None
    unprefixed_lines = ["".join(split_line[idx:]) for split_line in split_lines]
    return "".join(unprefixed_lines)



import enum
import inspect
import marko
import sys
from io import FileIO
from functools import cache
from marko.md_renderer import MarkdownRenderer
from pathlib import Path
from typing import Any, Generator, TypeAlias, Optional, Callable

from ..types import (
    RecordID, Image, Metadata, URI, Record, ValidationError
)


class ResourceType(enum.Enum):
    Generic = "generic"
    Code = "code"
    Document = "document"
    Documentation = "documentation"
    Example = "example"
    Image = "image"


def _is_resource(obj: Any):
    return bool(
        inspect.isclass(obj)
        and issubclass(obj, Resource)
    )


@cache
def get_resource_cls(resource_type: ResourceType):
    current_module = sys.modules[__name__]
    for _cls_name, cls in inspect.getmembers(current_module, predicate=_is_resource):
        if getattr(cls, "resource_type", None) == resource_type:
            return cls


ResourceContentType: TypeAlias = str | bytes | Image

class Resource:
    """

    """
    resource_type: ResourceType = ResourceType.Generic
    default_partition: str = "default"

    uri: URI
    id: RecordID | None = None
    content: ResourceContentType | None = None
    file_handle: FileIO | None = None
    metadata: Metadata |  None = None
    basedir: str | Path = ""
    content_is_complete: bool | None = None
    validated: bool

    def __init__(self,
        uri: URI | str,
        id: RecordID | None = None,
        content: str | bytes | Image | None = None,
        file_handle: FileIO | None = None,
        metadata: Metadata  |  None = None,
        basedir: str | Path = "",
        content_is_complete: bool | None = None,
    ) -> None:
        self.uri = URI(uri)
        self.id = id
        self.content = content
        self.file_handle = file_handle
        self.metadata = metadata
        self.basedir = basedir
        self.content_is_complete = content_is_complete
        self.validated = False

        if self.content is None and self.file_handle is None:
            raise ValueError("Either content or file_handle must be provided when defining a resource.")

    def validate(self, content: ResourceContentType | None) -> None:
        """
        Should raise a ValueError if resource is not valid
        """
        pass

    def read(self):
        from ..loaders.schemes import read_from_uri
        if self.content_is_complete and self.content is not None:
            return self.content
        elif self.file_handle and not self.file_handle.closed:
            try:
                self.file_handle.seek(0)
            except IOError:
                pass
            return self.file_handle.read()
        elif self.uri:
            return read_from_uri(self.uri)
        else:
            return None

    def as_records(self, splitter=None) -> Generator[Record, None, None]:
        content = self.read()

        if content is None:
            raise ValueError(f"Unable to determine content for resource `{self.id}` @ `{self.uri}`")

        # Validate before use if needed
        if not self.validated:
            self.validate(content=content)

        if splitter is None:
            chunks = [content]
        else:
            chunks = splitter.split_text(content)

        for chunk_num, chunk in enumerate(
                chunks,
                start=1,
        ):
            record = Record(
                id=f"{self.id}:{chunk_num}",
                uri=self.uri,
                metadata=self.metadata,
                content=chunk,
            )
            yield record


class DocumentResource(Resource):
    resource_type = ResourceType.Document
    default_partition: str = "default"


class DocumentationResource(DocumentResource):
    resource_type = ResourceType.Documentation
    default_partition: str = "documentation"


class CodeResource(DocumentResource):
    resource_type = ResourceType.Code
    default_partition: str = "code"


class ExampleResource(Resource):
    resource_type = ResourceType.Example
    default_partition: str = "examples"

    def validate(self, content: ResourceContentType | None = None):
        if content is None:
            content = self.read()
        try:
            tree = marko.Parser().parse(content)
        except Exception as err:
            raise ValueError(
                f"Example file at `{self.uri}` does not seem to be valid Example markdown file."
            ) from err

        sections = {}

        current_section = None
        chunk = []
        for node in tree.children:
            if node.get_type(snake_case=True) == "heading":
                if current_section:
                    sections[current_section] = chunk
                    chunk = []
                current_section = MarkdownRenderer().render_children(node).strip()
            elif node.get_type(snake_case=True) == "blank_line":
                continue
            elif current_section:
                chunk.append(node)
        if current_section:
            sections[current_section] = chunk

        # Remove any empty sections that may have been defined but have no content
        for key in sections.keys():
            if len(sections[key]) == 0:
                del sections[key]

        if not sections:
            # TODO: Include link to docs once those exist.
            raise ValidationError(f"Example file at `{self.uri}` is empty or does not meet the expected format.")
        if "Description" not in sections:
            raise ValidationError(f"Example file at `{self.uri}` is missing the required 'Description' section.")
        self.validated = True


ResourceFilter: TypeAlias = Optional[Callable[[Resource], bool]]


def isCodeResource(resource: Resource) -> bool:
    return isinstance(resource, CodeResource)


def isDocumentResource(resource: Resource) -> bool:
    return isinstance(resource, DocumentResource)


def isDocumentationResource(resource: Resource) -> bool:
    return isinstance(resource, DocumentationResource)


def isExampleResource(resource: Resource) -> bool:
    return isinstance(resource, ExampleResource)


def isImageResource(resource: Resource) -> bool:
    return isinstance(resource, Resource) and resource.resource_type == ResourceType.Image



# -*- coding: utf-8 -*-
import asyncio
import json
import logging
import re
import typing

from archytas.tool_utils import AgentRef, LoopControllerRef, tool
from beaker_kernel.lib.agent import BaseAgent

from beaker_bunsen.corpus.types import QueryResult

if typing.TYPE_CHECKING:
    from beaker_kernel.lib.context import BaseContext
    from .bunsen_context import BunsenContext


logger = logging.getLogger(__name__)


class BunsenAgent(BaseAgent):

    EXAMPLE_INTRO = """
Below are some similar examples of code that may be similar or related to the current request.
If the request from the user is similar enough to one of these examples, use it to help write code to answer the user's request.
""".strip()
    EXAMPLE_OUTRO = """
""".strip()

    CODE_PROMPT_INTRO = """
Please generate {self.context.subkernel.DISPLAY_NAME} code intended to run in Jupyter kernel "{self.context.subkernel.KERNEL_NAME}".
to satisfy the user's request below.
Note that this tool should generate the code directly from the LLM and not pick a tool to generate the code. This is the
code generation tool.
""".strip()
    CODE_PROMPT_OUTRO = """
Please generate the code as if you were programming inside a Jupyter Notebook and the code is to be executed inside a cell.
You MUST wrap the code with a line containing three backticks (```) before and after the generated code.
No addtional text is needed in the response, just the code block.
""".strip()


    context: "BunsenContext"

    def __init__(
        self,
        context: "BaseContext" = None,
        tools: list = None,
        **kwargs,
    ):
        self.context = context

        super().__init__(
            context=context,
            tools=tools,
            **kwargs
        )

    async def format_examples(self, examples: list[QueryResult]) -> str:
        code_example_str = "\n\n".join(
            """
======== example {num}: {example_id} start ========
{example}
======== example {num}: {example_id} end   ========
            """.strip().format(example_id=example["record"].id, example=example["record"].content, num=num)
            for num, example in enumerate(examples, start=1)
        )
        return "\n".join([
            self.EXAMPLE_INTRO,
            code_example_str,
            self.EXAMPLE_OUTRO,
        ])


    @tool()
    async def generate_code(self, code_request: str, agent: AgentRef, loop: LoopControllerRef):
        """
        Generated code to be run in an interactive Jupyter notebook.

        Input is a full grammatically correct question about or request for an action to be performed in the current environment.
        If you need more information on how to accomplish the request, you should use the other tools prior to using this one.

        Args:
            code_request (str): A fully grammatically correct set of instructions on the code that needs to be generated.
        """

        try:
            example_future = self.context.get_examples(
                query=code_request
            )
            state_future = self.context.get_subkerkel_state_description()

            examples, state_desc = await asyncio.gather(
                example_future,
                state_future
            )

            request_prompt = f"""
User's Request:
```
{code_request}
```
""".strip()

            prompt = [
                self.CODE_PROMPT_INTRO.format(self=self),
                request_prompt,
                self.context.library_description,
            ]
            if examples:
                prompt.append(await self.format_examples(examples))
            if state_desc:
                prompt.append(state_desc)
            prompt.append(self.CODE_PROMPT_OUTRO.format(self=self))

            code_generation_prompt = "\n\n".join(prompt)
            self.context.beaker_kernel.log("bunsen_prompt_code_gen", code_generation_prompt)
            response = await agent.inspect(code_generation_prompt)
            preamble, code, coda = re.split("```\w*", response)
            loop.set_state(loop.STOP_SUCCESS)

            result = {
                "action": "code_cell",
                "language": self.context.subkernel.SLUG,
                "content": code.strip(),
            }

            return json.dumps(result)
        except Exception as err:
            logger.error(err.args, exc_info=err)

    @tool()
    async def get_source_code(self, asset_type: str, asset_name: str, agent: AgentRef, loop: LoopControllerRef) -> list[str]:
        """
        Retrieves source code for a module, class, or function code asset.

        Input is a distinct identifier that uniquely distinguishes which asset you are wanting the source code of.
        Note that as this searches over the vectorized embeddings of the code using a nearest-neighbor search it is possible
        that you will receive irrelevent results if there is not a good match found.

        Args:
            asset_type (str): The type of asset you are looking up. Should be: "module", "class", or "function".
            asset_name (str): A distinct identifier that uniquely distinguishes which asset you are wanting the source code of.

        Returns:
            list: A list of chunks (strings) of source code that should correspond with the indicated asset. Order by cosine distance.
        """

        query = f"Definition of {asset_type} {asset_name}"
        matches = agent.context.get_source_code(
            query
        )
        return matches



from typing import TYPE_CHECKING
from ..types import DefaultType, Default

from ..util.splitters import TextSplitter, RecursiveCharacterTextSplitter


if TYPE_CHECKING:
    from .resource import Resource


def find_splitter_for_resource(
    resource: "Resource",
    chunk_size: int | DefaultType = 2000,
    chunk_overlap: int | DefaultType = 100,
) -> TextSplitter | None:
    from .resource import CodeResource, DocumentationResource, DocumentResource, ExampleResource, ResourceType
    if chunk_size is Default:
        chunk_size = 2000
    if chunk_overlap is Default:
        chunk_overlap = 100

    # Images are never chunked. Returning None indicates to not split.
    if resource.resource_type == ResourceType.Image:
        return None

    # Examples are currently not chunked, but may be required in the future if examples grow very large.
    # In such a case, we will probably want to rethink how we use examples.
    if isinstance(resource, ExampleResource):
        return None

    # Check extension for pre-mapped recursive splitter
    if '.' in resource.uri:
        _, extension = resource.uri.rsplit('.', maxsplit=1)
        if extension in RecursiveCharacterTextSplitter.LANGUAGES_BY_EXTENSION:
            return RecursiveCharacterTextSplitter.from_extension(
                extension=extension,
                chunk_size=chunk_size,
                chunk_overlap=chunk_overlap
            )

    # For code and documents of unprepared type, use default chunking
    if isinstance(resource, (CodeResource, DocumentResource, DocumentationResource)):
        return RecursiveCharacterTextSplitter(
                chunk_size=chunk_size,
                chunk_overlap=chunk_overlap
        )

    # Finally, the default
    return RecursiveCharacterTextSplitter(
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
    )



from .resource import (
    Resource, ResourceType, CodeResource, DocumentResource, DocumentationResource, ExampleResource, get_resource_cls,
    ResourceFilter, isCodeResource, isDocumentResource, isDocumentationResource, isExampleResource, isImageResource
)
from .splitting import find_splitter_for_resource


__all__ = [
    "Resource",
    "ResourceType",
    "ResourceFilter",
    "CodeResource",
    "DocumentResource",
    "DocumentationResource",
    "ExampleResource",
    "get_resource_cls",
    "isCodeResource",
    "isDocumentResource",
    "isDocumentationResource",
    "isExampleResource",
    "isImageResource",
    "find_splitter_for_resource",
]



import contextlib
import importlib
import importlib.util
import logging
import os
import pkgutil
import requests
import sys
import tarfile
import tempfile
from collections import deque, defaultdict
from pathlib import Path
from typing import Any

from .base import BaseLoader
from .schemes import LocalFileScheme, PythonModuleScheme, RCranScheme, read_from_uri
from ..types import Default, DefaultType, URI
from ..resources import CodeResource, ExampleResource, DocumentationResource, ResourceFilter

logger = logging.getLogger("beaker_bunsen")

class BaseCodeLoader(BaseLoader):

    Scheme = LocalFileScheme

    def read(
            self,
            location: str,
            base: str = ""
        ):
        if location.startswith(self.Scheme.URI_SCHEME):
            uri = location
        else:
            uri = self.Scheme.get_uri_for_location(location, base)
        return read_from_uri(uri)


class PythonLibraryLoader(BaseCodeLoader):

    Scheme = PythonModuleScheme
    SLUG = "python"

    def discover(
        self,
        locations: list[str] | DefaultType = Default,
        metadata: dict | DefaultType = Default,
        exclusions: list[str] | DefaultType = Default,
        filter: ResourceFilter | DefaultType = Default,
    ):
        """
        The `locations` should be Python packages, modules or submodules that are installed via the pyproject
        requirements.  Each location must be importable using the syntax `import {location}`, so you can include
        either top-level modules such as 'pandas' or submodules such as `pandas.plotting`.  The loader will recurse
        through and to include anything defined "below" the specified location in the import tree. I.e. location
        `pandas` will load all submodules, but `pandas.api.extensions` will not import any other submodules under
        `pandas.api.*` besides what is specified.

        Note: "from" import syntax is not valid. Loading a module loads the entire module and it's submodules and
        it is not possible to select only certain items from the module.

        Example: locations=["requests", "os.path", "pandas.core", "pandas.util"]
        """
        if locations is Default:
            locations = self.locations
        if metadata is Default:
            metadata = self.metadata
        if exclusions is Default:
            exclusions = self.exclusions
        if filter is Default:
            filter = self.filter

        # Update or define locations and exclusions based on '!' prefix in location.
        locations, parsed_exclusions = self.parse_locations(locations)
        exclusions.extend(parsed_exclusions)

        modules_to_collect = deque()
        for module_name in locations:
            uri = URI(module_name)
            if uri.scheme:
                module_name = uri.path
            module_spec = importlib.util.find_spec(module_name)
            if module_spec is None:
                raise ValueError(
                    f"Module '{module_name}' is not able to be imported. Please ensure that it is listed as a "
                    f"requirement and that 'require-runtime-dependencies' is enabled if error encountered during build."
                )
            modules_to_collect.append(module_spec)

        while modules_to_collect:
            module_spec = modules_to_collect.popleft()

            if not module_spec.origin or not module_spec.origin.endswith('.py'):
                logger.info(f"Skipping importing non-python file {module_spec.origin}")
                continue

            if getattr(module_spec, "submodule_search_locations", []):
                subpkg_info: pkgutil.ModuleInfo = pkgutil.iter_modules(path=module_spec.submodule_search_locations)
                subpkg_specs = (info.module_finder.find_spec(f"{module_spec.name}.{info.name}") for info in subpkg_info)
                if self.exclusions:
                    subpkg_specs = (
                        spec for spec in subpkg_specs
                        if not self.should_exclude(spec.name)
                    )
                modules_to_collect.extend(
                    subpkg_specs
                )

            if hasattr(module_spec, "loader"):
                source = module_spec.loader.get_source(module_spec.name)
            else:
                source = None

            origin_path = Path(module_spec.origin)
            for sys_path in sys.path:
                if origin_path.is_relative_to(sys_path):
                    basedir = sys_path
                    break
            else:
                basedir = ""

            resource = CodeResource(
                uri=self.Scheme.get_uri_for_location(module_spec.name),
                content=source,
                metadata={
                    "package": module_spec.name,
                    "type": "code",
                    **metadata,
                }
                # basedir=basedir
            )
            resource.id = self.get_id_for_resource(resource)
            if filter and not filter(resource):
                continue
            yield resource


class RCRANLocalCache(contextlib.AbstractContextManager):
    # Class attributes
    remote_package_cache: dict[str, dict[str, str]] = {}
    local_package_cache: dict[str, str] = {}
    tempdir_context_holder: dict[str, contextlib.AbstractContextManager] = {}
    ref_counts: dict[str, int] = defaultdict(lambda: 0)

    # Instance/context attributes
    repo: str
    context_locations: list[str]

    def __init__(
        self,
        locations: list[str],
        repo="https://cran.rstudio.com/src/contrib",
    ) -> None:
        self.repo = repo
        self.context_locations = [
            URI(location).path if URI(location).scheme == "rcran-package" else location for location in locations
        ]
        if not self.remote_package_cache:
            self.build_package_cache()

    def __enter__(self) -> dict[str, str]:
        results = {}
        for location in self.context_locations:
            # Increase ref counter to reflect usage in new context
            self.__class__.ref_counts[location] += 1

            # Return the cached if it exists
            existing_cache = self.local_package_cache.get(location, None)
            if existing_cache and os.path.isdir(existing_cache):
                results[location] = existing_cache
                continue

            if '@' in location:
                package_name, version = location.split("@", maxsplit=1)
            else:
                package = self.remote_package_cache.get(location, None)
                if not package:
                    raise LookupError(f"Unable to find CRAN package name '{location}'")
                package_name = package["package"]
                version = package["version"]

            source_tarball_url = f"{self.repo}/{package_name}_{version}.tar.gz"
            tarball_req = requests.get(source_tarball_url, stream=True)

            context_mgr = tempfile.TemporaryDirectory()
            tmpdir = context_mgr.__enter__()
            results[location] = tmpdir
            self.tempdir_context_holder[location] = context_mgr
            self.local_package_cache[location] = tmpdir

            tar_file = tarfile.open(mode="r:gz", fileobj=tarball_req.raw)
            tar_file.extractall(path=tmpdir)
        return results

    def __exit__(self, exc_type, exc_value, traceback):
        # Update ref counts for local context
        for location in self.context_locations:
            self.__class__.ref_counts[location] -= 1
        # Clean up if there are no references
        locations_to_cleanup = [
            location
            for location, ref_count in self.__class__.ref_counts.items()
            if ref_count == 0
        ]
        for location in locations_to_cleanup:
            context_mgr = self.tempdir_context_holder[location]
            context_mgr.__exit__(exc_type, exc_value, traceback)
            del self.tempdir_context_holder[location]
            del self.local_package_cache[location]
            del self.__class__.ref_counts[location]
        return super().__exit__(exc_type, exc_value, traceback)

    def build_package_cache(self):
        package_page = f"{self.repo}/PACKAGES"
        package_req = requests.get(package_page)
        package_content = package_req.text
        packages = {}
        current_package = {}
        for line in package_content.splitlines():
            if line == "":
                packages[current_package["package"]] = current_package
                current_package = {}
                continue
            elif line.startswith("       "):
                current_package[label] += " " + line.lstrip()
                continue
            try:
                label, value = line.split(":", maxsplit=1)
                label = label.lower().strip()
                value = value.strip()
            except ValueError:
                raise
            current_package[label] = value
        self.remote_package_cache.clear()
        self.remote_package_cache.update(packages)


class RCRANSourceLoader(BaseCodeLoader):
    REPO: str = "https://cran.rstudio.com/src/contrib"
    SLUG: str = "rcran"

    Scheme = RCranScheme

    remote_package_cache: dict[str, Any] | None = {}
    local_package_cache: dict[str, str] | None = {}

    def discover(
        self,
        locations: list[str] | DefaultType = Default,
        metadata: dict | DefaultType = Default,
        exclusions: list[str] | DefaultType = Default,
        filter: ResourceFilter | DefaultType = Default,
    ):
        """
        The `locations` should be R packages, as they are named in Cran with an option version number separated from the
        package name by an @ symbol.

        Example: locations=["purr", "jsonlite", "shiny@1.8.1", "leaflet@2.0"]
        """
        if locations is Default:
            locations = self.locations
        if metadata is Default:
            metadata = self.metadata
        if exclusions is Default:
            exclusions = self.exclusions
        if filter is Default:
            filter = self.filter

        # Update or define locations and exclusions based on '!' prefix in location.
        locations, parsed_exclusions = self.parse_locations(locations)
        exclusions.extend(parsed_exclusions)

        with RCRANLocalCache(locations=locations) as cache:
            for package_name, tempdir in cache.items():
                for dirpath, _dirnames, filenames in os.walk(top=tempdir):
                    for filename in filenames:
                        if filename.startswith('.'):
                            continue
                        full_path = Path(os.path.join(dirpath, filename))
                        if "R" in full_path.parts and full_path.suffix == ".R":
                            resource_type = CodeResource
                            location = str(full_path.relative_to(tempdir))
                        elif "vignettes" in full_path.parts and full_path.suffix.startswith(".R"):
                            resource_type = ExampleResource
                            location = str(full_path.relative_to(tempdir))
                        elif "man" in full_path.parts and full_path.suffix.startswith(".R"):
                            resource_type = DocumentationResource
                            location = str(full_path.relative_to(tempdir))
                        else:
                            continue

                        with open(full_path, 'r') as fh:
                            content = fh.read()
                            fh.seek(0)
                            resource = resource_type(
                                uri=self.Scheme.get_uri_for_location(
                                    base=package_name,
                                    location=location,
                                ),
                                file_handle=fh,
                                content=content,
                                metadata={
                                    "package": package_name,
                                    "path": str(full_path),
                                    "type": resource_type.resource_type.value,
                                    **metadata,
                                }
                            )
                            resource.id = self.get_id_for_resource(resource)
                            if filter and not filter(resource):
                                continue
                            yield resource
